__version__ = '1.6.9'

from .game import Game, LocalGame, RemoteGame
from .player import Player, load_player
from .types import *
from .utilities import is_illegal_castle, is_psuedo_legal_castle, ChessJSONEncoder, ChessJSONDecoder
from .play import play_local_game, play_remote_game, play_turn, notify_opponent_move_results, play_sense, play_move
from .history import Turn, GameHistory, GameHistoryEncoder, GameHistoryDecoder
import chess
